#include "Shooter.h"
#include "ShooterPID.h"
// Panel - Green
// Line - Yellow
// Port - Blue
// Target Corner - Red
const double kHoodMin = .48; // .48
const double kHoodMax = kHoodMin + .486; 
//these need to be modified for new challenges
const double kGreenZone = kHoodMin + .34; //.81 Green Zone
const double kYellowZone = kHoodMin + .457; //.937 Yellow Zone
const double kBlueZone = kHoodMin + .478; // .958 Blue Zone
const double kRedZone = kHoodMin + .483; // .963 Red Zone


//dont touch these
const double kServoStopped = .5; 
const double kServoForward = .7; 
const double kServoBackward = .3; 
const double kHoodTolerance = .01;

const double kVelocityScaleFactor = .5; // 2:1 gear ratio

const double kReadyToShootDeadzone = .2;

const double kPollCooldown = .05;
const double kHistoryThreshold = .5;
const double kVelocityThreshold = .5;
const double kMovingDeadzone = 100;

const double kTransitionFeedSpeed = 1;

const double kAverageTolerance = 200; // 100 rpm tolerance

const int kRPMGreenZone = 3500; // Green Zone 4000?
const int kRPMRedZone = 5000; // Red Zone
const int kRPMOtherZones = 4800; // Yellow Red and Blue Zones 80% Power

const int kNeo550MaxRPM = 916;

Shooter::Shooter(bool rapidFire) : primerMotor(ThunderSparkMax::create(ThunderSparkMax::MotorID::ShooterPrimer)),
                     shooterMotorMaster(ThunderSparkMax::create(ThunderSparkMax::MotorID::ShooterRight)),
                     shooterMotorSlave(ThunderSparkMax::create(ThunderSparkMax::MotorID::ShooterLeft)),
                     transitionMotor(ThunderSparkMax::create(ThunderSparkMax::MotorID::StorageTransitionToShooter)),
                     pidController(shooterMotorMaster->getSparkPID())
{
    shooterMotorMaster->SetIdleMode(ThunderSparkMax::IdleMode::COAST);

    shooterMotorMaster->SetInverted(true);

    shooterMotorSlave->SetIdleMode(ThunderSparkMax::IdleMode::COAST);
    shooterMotorSlave->Follow(shooterMotorMaster, true);

    cooldown.Start();

    rapidFiring = rapidFire;

    primerMotor->SetInverted(true);

    for(int i = 0; i < kHistoryCount; i++) {
        velocityHistory[i] = 0;
    }

    pidController->SetP(ShooterPID::kP);
    pidController->SetI(ShooterPID::kI);
    pidController->SetD(ShooterPID::kD);
    pidController->SetFF(ShooterPID::kF);
    pidController->SetIZone(ShooterPID::kIZone);
    pidController->SetOutputRange(ShooterPID::kMinOutput, ShooterPID::kMaxOutput);
}

void Shooter::reset()
{ 
    primerMotor->Set(0);
    shooterMotorMaster->Set(0);
    transitionMotor->Set(0);
    pidController->SetReference(0, rev::ControlType::kVelocity);
}

void Shooter::debug(Feedback *feedback)
{
    feedback->sendDouble("shooter", "velocity", shooterMotorMaster->GetVelocity());
    feedback->sendDouble("shooter", "output master", shooterMotorMaster->Get());
    feedback->sendDouble("shooter", "output slave", shooterMotorSlave->Get());
    feedback->sendBoolean("shooter", "want to spin", wantToSpin);
    feedback->sendBoolean("shooter", "want to shoot", wantToShoot);
    feedback->sendDouble("shooter", "hood angle", reverseHoodEncoder(hoodPotentiometer.Get()));
    feedback->sendDouble("shooter", "manual hood speed", hoodSpeedManual);

    const char* presetName = "";
    switch(hoodPos) {
        case GREEN:
        presetName = "green";
        break;
        case YELLOW:
        presetName = "yellow";
        break;
        case BLUE:
        presetName = "blue";
        break;
        case RED:
        presetName = "red";
        break;
        case MANUAL:
        presetName = "manual";
        break;
    }
    feedback->sendString("shooter", "preset", presetName);
    feedback->sendDouble("shooter", "hood position", hoodServo.Get());

    feedback->sendDouble("shooter", "set point", setPoint);

    if(manualControl) {
        feedback->sendDouble("thunderdashboard", "shooter_speed", 100*desiredSpeed);
    }
    else {
        feedback->sendDouble("thunderdashboard", "shooter_speed", 100*(shooterMotorMaster->GetVelocity() / ShooterPID::kMaxRPM));
    }

    feedback->sendDouble("thunderdashboard", "shooter_hood", hoodServo.Get() * 100);

    feedback->sendDouble("shooter", "manual speed", desiredSpeed);
    feedback->sendBoolean("shooter", "manual enabled", manualControl);

    if(initDebug) {
        initDebug = false;

        feedback->sendEditableDouble("shooter", "pid f", ShooterPID::kF);
        feedback->sendEditableDouble("shooter", "pid p", ShooterPID::kP);
    }
    else {
        if(feedback->getEditableDouble("shooter", "pid f", ShooterPID::kF) != ShooterPID::kF) {
            pidController->SetFF(feedback->getEditableDouble("shooter", "pid f", ShooterPID::kF));
        }

        if(feedback->getEditableDouble("shooter", "pid p", ShooterPID::kP) != ShooterPID::kP) {
            pidController->SetP(feedback->getEditableDouble("shooter", "pid p", ShooterPID::kP));
        }
    }

    feedback->sendDouble("shooter", "primer output", primerMotor->GetVelocity() / kNeo550MaxRPM);


    //dashboard graphing

    feedback->sendString("shooter", "chart_plotmin", "0"); // min
    feedback->sendString("shooter", "chart_plotmax", "0"); // max
    feedback->sendString("shooter", "chart_platlab", "RPM"); // label
    feedback->sendString("shooter", "chart_plotctr", "0"); // centerline

    feedback->sendString("shooter", "chart_plot", "%d", shooterMotorMaster->GetVelocity());


}
void Shooter::process()
{
    double finalTransitionSpeed = 0;
    double finalPrimerSpeed = 0;
    double shooterSetPoint = 0;

    if(wantToSpin) {
            switch(hoodPos) {
                case GREEN:
                shooterSetPoint = kRPMGreenZone;
                break;
                case YELLOW:
                shooterSetPoint = kRPMOtherZones;
                break;
                case BLUE:
                shooterSetPoint = kRPMOtherZones;
                break;
                case RED:
                shooterSetPoint = kRPMRedZone;
                break;
                case MANUAL:
                shooterSetPoint = kRPMOtherZones;
                break;
            }

        

        finalPrimerSpeed = 1;
        if(wantToShoot && isAtVelocity()) {
            finalTransitionSpeed = kTransitionFeedSpeed;
        }
    }
    else {

    }


    transitionMotor->Set(finalTransitionSpeed);
    primerMotor->Set(finalPrimerSpeed);

    if(manualControl && wantToSpin) {
        pidController->SetReference(desiredSpeed * ShooterPID::kMaxRPM, rev::ControlType::kVelocity);
    }
    else {
        pidController->SetReference(shooterSetPoint, rev::ControlType::kVelocity);
    }
    setPoint = shooterSetPoint;
    

    if(cooldown.Get() > kPollCooldown) {
        cooldown.Reset();
        shiftVelocity(shooterMotorMaster->GetVelocity());
    }

    double servoSpeed = kServoStopped;

    double hoodSensor = reverseHoodEncoder(hoodPotentiometer.Get());

    switch (hoodPos)
    {

    case YELLOW:
        if(hoodSensor+kHoodTolerance >= kYellowZone && hoodSensor-kHoodTolerance <= kYellowZone){
            servoSpeed = kServoStopped;
        }
        else if(hoodSensor-kHoodTolerance<kYellowZone){
            servoSpeed = kServoForward;
        }
        else if(hoodSensor+kHoodTolerance>kYellowZone){
            servoSpeed = kServoBackward;

        }
        else{
            servoSpeed = kServoStopped;
        }
        break;
    case RED:
        if(hoodSensor+kHoodTolerance >= kRedZone && hoodSensor-kHoodTolerance <= kRedZone){
            servoSpeed = kServoStopped;
        }
        else if(hoodSensor-kHoodTolerance<kRedZone){
 
            servoSpeed = kServoForward;
        }
        else if(hoodSensor+kHoodTolerance>kRedZone){
           
            servoSpeed = kServoBackward;
        }
        else{
            
            servoSpeed = kServoStopped;
        }
        break;
    case GREEN:
        if(hoodSensor+kHoodTolerance>=kGreenZone && hoodSensor-kHoodTolerance<=kGreenZone){
 
            servoSpeed = kServoStopped;
        }
        else if(hoodSensor-kHoodTolerance<kGreenZone){
      
            servoSpeed = kServoForward;
        }
        else if(hoodSensor+kHoodTolerance>kGreenZone){
      
            servoSpeed = kServoBackward;
        }
        else{

            servoSpeed = kServoStopped;
        }
        break;
    case BLUE:
        if(hoodSensor+kHoodTolerance>=kBlueZone && hoodSensor-kHoodTolerance <=kBlueZone){
      
            servoSpeed = kServoStopped;
        }
        else if(hoodSensor-kHoodTolerance<kBlueZone){
     
            servoSpeed = kServoForward;
        }
        else if(hoodSensor+kHoodTolerance>kBlueZone){
    
            servoSpeed = kServoBackward;
        }
        else{
            servoSpeed = kServoStopped;
        }
        break;
    case MANUAL:
        servoSpeed = hoodSpeedManual;
        break;

    }
    /*potentiometer values are between -1 and 1, servo values are 0 to 1. 
        i have no idea what the actual values are so these are just placeholders*/
    if(hoodSensor > kHoodMax && servoSpeed > .5){
        servoSpeed = kServoStopped;
    }
    else if(hoodSensor < kHoodMin && servoSpeed < .5){
        servoSpeed = kServoStopped;
    }
        //the math here inverts the servo range, servo spins opposite
        hoodServo.Set(((-1*(servoSpeed-.5))+.5));
        //hoodServo.Set(servoSpeed);
        //hoodSensor = (-1*(hoodPotentiometer.Get()-1));
}

void Shooter::setShooterSpinup(bool shouldSpin)
{
    if(wantToSpin == false && shouldSpin == true) {
        for(int i = 0; i < kHistoryCount; i++) {
            shiftVelocity(0);
        }
    }
    wantToSpin = shouldSpin;
    
}
void Shooter::feedPowerCell(bool feed)
{
    wantToShoot = feed;
}

void Shooter::setShooterSpeed(double speed)
{
    desiredSpeed = fabs(speed);
}


void Shooter::hood(HoodPosition position)
{
    hoodPos = position;
}

bool Shooter::isAtVelocity() {
    if(rapidFiring) {
        return true;
    }
    else {
        return shooterMotorMaster->GetVelocity() > setPoint - kAverageTolerance;
    }
}

void Shooter::shiftVelocity(double velocity) {
    for(int i = 0; i < kHistoryCount-1; i++) {
        velocityHistory[i] = velocityHistory[i+1];
    }
    velocityHistory[kHistoryCount-1] = velocity;
}

void Shooter::enableManualControl(bool enabled) {
    manualControl = enabled;
}

double Shooter::reverseHoodEncoder(double value) {
    value = (1-value);
    return value;
}

void Shooter::hoodManual(double hoodSpeed) {
    if(hoodSpeedManual != kServoStopped){
        hoodPos = MANUAL;
    } 
    hoodSpeedManual = hoodSpeed;
}
