#pragma once

#include <frc/Encoder.h>
#include <frc/AnalogGyro.h>
#include <frc/ADXRS450_Gyro.h>
#include "IOMap.h"
#include <rev/SparkMax.h>
#include "Feedback.h"
#include "Lights.h"
#include "ThunderSparkMax.h"

class Drive {
public:
    Drive(Lights *lights);

    void process();
    void reset();

    /**
     * Sets the speed of the left and right sides of the drive train. 
     * Input values range from -1.0 and 1.0
     * 
     * Less than 0: reverse
     * 
     * Equal to 0: stopped
     * 
     * Greater than 0: forward
     */
    void setSpeed(double leftSpeed, double rightSpeed);

    /**
     * Drives given distance in inches
     * Speed is absolute, so to drive in reverse distance must be negative. Ranges from 0.0 to 1.0
     */
    void autoForward(double distance, double speed);

    /**
     * Turns given angle in degrees, positive turning the robot right.
     * Speed is absolute, so it has no impact on direction. Ranges from 0 to 1.0.
     */
    void autoPivot(double angle, double speed);

    /**
     * Returns status of the last issued auto action
     * true if complete
     * false if not complete
     */
    bool autoComplete();

    /**
     * prints debug strings
     */
    void debug(Feedback* feedback);

    /**
     * Calibrates gyro
     */
    void calibrateGyro();

    /*
    Stops using encoders to backdrive/brake
    */
    void setBackdriveBroken(bool broken);
private:
    ThunderSparkMax *motorLeftFront;
    ThunderSparkMax *motorLeftRear;
    ThunderSparkMax *motorRightFront;
    ThunderSparkMax *motorRightRear;

    frc::Encoder encLeft{DIO_DRIVE_ENC_LEFT_A, DIO_DRIVE_ENC_LEFT_B};
    frc::Encoder encRight{DIO_DRIVE_ENC_RIGHT_B, DIO_DRIVE_ENC_RIGHT_A};   

#ifdef TEST_BOARD
    frc::ADXRS450_Gyro gyro;
#else
    frc::AnalogGyro gyro{ANALOG_GYRO};
#endif
    bool gyroCalibrated = false; 

    enum DriveState {
        STOPPED, AUTO_STRAIGHT, AUTO_PIVOT, MANUAL
    };

    DriveState driveState = STOPPED;

    static const int kCircularHistorySize = 10;
    double circularHistoryLeft[kCircularHistorySize];
    double circularHistoryRight[kCircularHistorySize];

    double manualSpeedLeft = 0;
    double manualSpeedRight = 0;

    double autoDistance = 0;
    double autoAngle = 0;
    double autoSpeed = 0;
    bool autoFinished = false;
    frc::Timer finishTimer{};
    double autoStartEncL = 0;
    double autoStartEncR = 0;
    frc::Timer skipTimer{};
    bool skippingAutoStep = false;

    frc::Timer gyroWarning{};
    bool gyroWasRed = false;

    double velocityLeft = 0;
    double velocityRight = 0;
    Lights *lights;
    void shiftHistory(double left, double right);
    double readEncLeft();
    double readEncRight();
    void resetEncoders();

    bool backdriveEncBroken = false;
};